﻿using System;
using System.Net;
using System.Configuration;
using System.Security;
using System.Security.Principal;
using System.Web;
using System.Web.Configuration; 
using System.Web.Security;
using System.Text;
using System.Diagnostics;
using System.Threading;
using AspNetMvcRestService.Security;

namespace AspNetMvcRestService
{
    #region AsyncAuthResult

    /// <summary>
    /// A class containing an IAsyncResult implementation
    /// that wraps a ManualResetEvent. Note this is  relatively
    /// simple implementation. A more complete (generic)
    /// implementation is available by Jeffrey Richter, found at
    /// http://msdn.microsoft.com/en-us/magazine/cc163467.aspx.
    /// My thanks to Jeff Prosise for the basis of this
    /// implementation.
    /// </summary>
    internal class AsyncAuthResult : IAsyncResult
    {
        private AsyncCallback _callback;
        private object _state;
        private ManualResetEvent _event;
        private bool _completed = false;
        private bool _result = false;
        private object _lock = new object();
        private Exception _ex = null;

        public AsyncAuthResult(AsyncCallback callback, object state)
        {
            _callback = callback;
            _state = state;
        }

        public Object AsyncState
        {
            get { return _state; }
        }

        public bool CompletedSynchronously
        {
            get { return false; }
        }

        public bool IsCompleted
        {
            get { return _completed; }
        }

        public WaitHandle AsyncWaitHandle
        {
            get
            {
                lock (_lock)
                {
                    if (_event == null)
                        _event = new ManualResetEvent(IsCompleted);
                    return _event;
                }
            }
        }

        public void CompleteCall(bool result, Exception exception)
        {
            // Passing null for exception means no error occurred
            _ex = exception;

            // Cache result
            _result = result;

            lock (_lock)
            {
                _completed = true;
                if (_event != null)
                    _event.Set();
            }

            if (_callback != null)
            {
                // Invoke callback...
                _callback(this);
            }
        }

        public bool EndInvoke()
        {
            // Operation is done...if an exception occured, throw it
            if (_ex != null) throw _ex;

            return _result;
        }
    }

    #endregion

    #region AuthenticationManager

    /// <summary>
    /// A class to manage the authentication and authorization
    /// workflow. Note for asynchronous operations, it spins
    /// up a new thread rather than consume one from the CLR thread
    /// pool. This is because we don't want to consume CLR thread
    /// pool threads. Instead, we leave those for ASP.NET to use
    /// to process incoming requests.
    /// </summary>
    internal sealed class AuthenticationManager
    {
        const string BasicAuthProtocol = "basic";
        HttpContext _context = null;

        public AuthenticationManager(HttpContext currentContext)
        {
            _context = currentContext;
        }

        public bool AuthenticateUser()
        {
            return InternalAuthenticateUser();
        }

        public IAsyncResult BeginAuthenticateUser(AsyncCallback callback, Object state)
        {
            // Create IAsyncResult object identifying the 
            // asynchronous operation.
            AsyncAuthResult ar = new AsyncAuthResult(callback, state);

            // Create a new thread to perform the operation...note we can't
            // use a thread from the thread pool as this defeats the purpose
            // of the asynchronous implementation (we're trying to leave
            // *those* threads alone for ASP.NET to use to service incoming
            // requests).
            Thread worker = new Thread(this.InternalAuthenticateUserAsync);
            worker.Start(ar);

            return ar;  // Return the IAsyncResult to the caller
        }

        public bool EndAuthenticateUser(IAsyncResult asyncResult)
        {
            // We know that it's really an AsyncAuthResult object
            AsyncAuthResult ar = asyncResult as AsyncAuthResult;
            return ar.EndInvoke();
        }

        private void InternalAuthenticateUserAsync(Object asyncResult)
        {
            // We know that it's really an AsyncAuthResult object
            AsyncAuthResult ar = asyncResult as AsyncAuthResult;

            // Perform the operation..if successful set the result. If
            // not, record the exception.
            try
            {
                bool authorized = this.InternalAuthenticateUser();
                ar.CompleteCall(authorized, null);
            }
            catch (Exception ex)
            {
                ar.CompleteCall(false, ex);
            }
        }

        private bool InternalAuthenticateUser()
        {
            // Pull the authorization information for this URI
            string url = _context.Request.Url.LocalPath;
            AuthorizationManager authMgr = AuthorizationManager.CreateAuthorizationManager(url);

            // Check to see if we need authentication
            if (authMgr.RequiresAuthorization)
            {
                // Look for the authorization header
                bool isAuthorized = false;
                string authHeader = _context.Request.Headers["Authorization"];
                if (!String.IsNullOrEmpty(authHeader) && authHeader.ToLower().StartsWith(BasicAuthProtocol))
                {
                    // Pull the credentials
                    string[] credentials = stripCredentials(authHeader);

                    // Authenticate
                    if (credentials.Length == 2 &&
                        Membership.ValidateUser(credentials[0], credentials[1]))
                    {
                        // Set the caller's credentials...note we may rescind this
                        // depending upon the authorization check from the configuration
                        // file.
                        _context.User = new ServiceUser(credentials[0], true);

                        // Verify the configuration file supports this user
                        if (authMgr.VerifyAuthorization(_context.User))
                        {
                            // Mark as authorized
                            isAuthorized = true;
                        }
                        else
                        {
                            // Rescind temporary authorization and indicate that
                            // the resource is forbidden.
                            _context.Response.StatusCode = (Int32)HttpStatusCode.Forbidden;

                            // End the request
                            return false;
                        }
                    }
                }

                // Check for authentication and return HTTP 401 if none
                if (!isAuthorized)
                {
                    // Assign the status code
                    _context.Response.StatusCode = (Int32)HttpStatusCode.Unauthorized;

                    // Calculate the realm
                    string realm = String.Format("Basic realm=\"{0}\"", _context.Request.Url.GetLeftPart(UriPartial.Path));

                    // Add the authentication header
                    _context.Response.AddHeader("WWW-Authenticate", realm);
                }
                return isAuthorized;
            }
            return true;
        }

        string[] stripCredentials(string authHeader)
        {
            // Remove the "basic"
            string rawCredentials = authHeader.Substring(6).Trim();

            // Remove Base64 encoding
            byte[] bytes = Convert.FromBase64String(rawCredentials);
            string credentials = UTF8Encoding.UTF8.GetString(bytes);

            // Return separated string array
            return credentials.Split(':');
        }
    }

    #endregion

    #region AuthorizationManager

    /// <summary>
    /// A class that wraps authorization, allowing us to easily
    /// determine if the authenticating user is a member of any
    /// specific role.
    /// </summary>
    internal sealed class AuthorizationManager
    {
        private CommaDelimitedStringCollection _allowedRoles = new CommaDelimitedStringCollection();
        private CommaDelimitedStringCollection _deniedRoles = new CommaDelimitedStringCollection();
        private CommaDelimitedStringCollection _allowedUsers = new CommaDelimitedStringCollection();
        private CommaDelimitedStringCollection _deniedUsers = new CommaDelimitedStringCollection();

        private AuthorizationManager()
        {
        }

        public CommaDelimitedStringCollection AllowedRoles
        {
            get { return _allowedRoles; }
        }

        public CommaDelimitedStringCollection DeniedRoles
        {
            get { return _deniedRoles; }
        }

        public CommaDelimitedStringCollection AllowedUsers
        {
            get { return _allowedUsers; }
        }

        public CommaDelimitedStringCollection DeniedUsers
        {
            get { return _deniedUsers; }
        }

        public bool HasRoles
        {
            get { return (AllowedRoles.Count > 0 || DeniedRoles.Count > 0); }
        }

        public bool HasNonDefaultUsers
        {
            get
            {
                // If there is only the default user indicated,
                // ignore that special case and return the
                // fact we have no other users than the
                // default.
                bool hasUsers = false;
                if (AllowedUsers.Count == 1 &&
                    AllowedUsers.Contains("*") &&
                    AllowedRoles.Count == 0)
                {
                    // Do nothing
                    ; 
                }
                else
                {    
                    // Check if we have allowed users
                    hasUsers |= AllowedUsers.Count > 0;

                    // Check if we have denied users
                    hasUsers |= DeniedUsers.Count > 0;
                }

                return hasUsers;
            }
        }

        public bool RequiresAuthorization
        {
            get { return (HasRoles || HasNonDefaultUsers); }
        }

        public bool VerifyAuthorization(IPrincipal user)
        {
            // check user against a set of business rules:
            //
            // 1) If the user is denied by username, deny the user
            // 2) If the user is allowed by default without roles, allow user
            // 3) If the user is denied by default without roles, deny user
            // 4) If the user is denied by default but allowed based on role
            //    membership, allow user
            bool userDeniedByName = DeniedUsers.Contains(user.Identity.Name);
            bool userAllowedByDefault = ((AllowedUsers.Contains(user.Identity.Name) || AllowedUsers.Contains("*")) && AllowedRoles.Count == 0);
            bool userDeniedByDefault = ((DeniedUsers.Contains("*") || AllowedUsers.Contains("?")) && AllowedRoles.Count == 0);
            bool userAllowedByRole = false;
            foreach (string role in AllowedRoles)
            {
                if (user.IsInRole(role))
                {
                    userAllowedByRole = true;
                    break;
                }
            }

            return (userDeniedByName || userAllowedByDefault || userDeniedByDefault || userAllowedByRole);
        }

        public static AuthorizationManager CreateAuthorizationManager(string url)
        {
            Configuration config =
            WebConfigurationManager.OpenWebConfiguration(url);
            AuthorizationSection configSection =
            (AuthorizationSection)config.GetSection("system.web/authorization");
            AuthorizationRuleCollection rules = configSection.Rules;

            AuthorizationManager configInfo = new AuthorizationManager();
            for (int i = 0; i < rules.Count; i++)
            {
                if (rules[i].Roles.Count > 0)
                {
                    if (rules[i].Action.ToString() == "Allow")
                        configInfo.AllowedRoles.AddRange(rules[i].Roles.ToString().Split(','));
                    else if (rules[i].Action.ToString() == "Deny")
                        configInfo.DeniedRoles.AddRange(rules[i].Roles.ToString().Split(','));
                }
                if (rules[i].Users.Count > 0)
                {
                    if (rules[i].Action.ToString() == "Allow")
                        configInfo.AllowedUsers.AddRange(rules[i].Users.ToString().Split(','));
                    else if (rules[i].Action.ToString() == "Deny")
                        configInfo.DeniedUsers.AddRange(rules[i].Users.ToString().Split(','));
                }
            }

            return configInfo;
        }
    }

    #endregion

    #region Synchronous authorization module implementation

    public class AuthModule : IHttpModule
    {
        /// <summary>
        /// You will need to configure this module in the web.config file of your
        /// web and register it with IIS before being able to use it. For more information
        /// see the following link: http://go.microsoft.com/?linkid=8101007
        /// 
        /// In addition, this module takes over the functions of the default
        /// UrlAuthorization module, so you'll need to remove it:
        /// 
        /// &lt;remove name="UrlAuthorization" /&gt;
        /// 
        /// For more information see the following link:
        /// http://technet.microsoft.com/en-us/library/cc731223.aspx
        /// </summary>
        #region IHttpModule Members

        public void Dispose()
        {
        }

        public void Init(HttpApplication context)
        {
            // Hook the authentication request event
            context.AuthenticateRequest += new EventHandler(OnAuthenticateRequest);
        }

        #endregion

        #region Event handlers

        void OnAuthenticateRequest(object sender, EventArgs e)
        {
            // Pull the current context
            HttpContext context = HttpContext.Current;

            // Authenticate
            AuthenticationManager authMgr = new AuthenticationManager(context);
            if (!authMgr.AuthenticateUser())
            {
                // End the request
                context.Response.End();
            }
        }

        #endregion
    }

    #endregion

    #region Asynchronous authorization module implementation

    /// <summary>
    /// You will need to configure this module in the web.config file of your
    /// web and register it with IIS before being able to use it. For more information
    /// see the following link: http://go.microsoft.com/?linkid=8101007
    /// 
    /// In addition, this module takes over the functions of the default
    /// UrlAuthorization module, so you'll need to remove it:
    /// 
    /// &lt;remove name="UrlAuthorization" /&gt;
    /// 
    /// For more information see the following link:
    /// http://technet.microsoft.com/en-us/library/cc731223.aspx
    /// </summary>
    public class AsyncAuthModule : IHttpModule
    {
        #region Fields

        // We *must* cache the context. When the "end" method is
        // invoked, a CLR thread will be executing (versus an IIS
        // thread), so there is no "current" context available
        // for the thread to use. That is, HttpContext.Current
        // will return null. We can cache the context in this manner
        // even though three different threads will eventually
        // use this object since the runtime suspends this module
        // invocation entirely and restarts it in the "end" 
        // method.
        HttpContext _context = null;

        // We'll cache this as well so we can call its "end"
        // method.
        AuthenticationManager _authMgr = null;

        #endregion

        #region IHttpModule Members

        public void Dispose()
        {
        }

        public void Init(HttpApplication application)
        {
            // Hook authentication
            application.AddOnAuthenticateRequestAsync(
                new BeginEventHandler(BeginAuthenticateRequestHandlerExecute),
                new EndEventHandler(EndAuthenticateRequestHandlerExecute)
            );
        }

        #endregion

        #region Event handlers

        IAsyncResult BeginAuthenticateRequestHandlerExecute(Object source,
            EventArgs e, AsyncCallback cb, Object state)
        {
            // Pull the current context
            // Pull the current context
            HttpApplication application = source as HttpApplication;
            _context = application.Context;

            // Create an instance of the authentication manager and invoke
            // its asynchronous method.
            _authMgr = new AuthenticationManager(_context);
            return _authMgr.BeginAuthenticateUser(cb, state);
        }

        void EndAuthenticateRequestHandlerExecute(IAsyncResult ar)
        {
            // Collect authentication/authorziation result...if true, the
            // user is good to go. If false, the context will contain the
            // appropriate response code and headers.
            bool authorized = _authMgr.EndAuthenticateUser(ar); // can throw...

            // If the context contains anything but "OK" we know
            // to terminate the request. The appropriate headers
            // will already be set up by the authentication manager
            // object.
            if (!authorized || _context.Response.StatusCode != (Int32)HttpStatusCode.OK)
            {
                // End the request
                _context.Response.End();
            }
        }

        #endregion
    }

    #endregion
}
